/*
Copyright 2020 Peter Howkins

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the
"Software"), to deal in the Software without restriction, including
without limitation the rights to use, copy, modify, merge, publish,
distribute, sublicense, and/or sell copies of the Software, and to
permit persons to whom the Software is furnished to do so, subject to
the following conditions:

The above copyright notice and this permission notice shall be included
in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/
#ifndef __OS_H__
#define __OS_H__

/**
 * @file
 * @brief BBC MOS API wrappers
 * 
 * C style wrappers around the BBC Micro OS API calls 
 */ 

#include <string.h>
#include "stdint.h"


/**
 * OSWRCH Write character to currently selected output stream
 *
 * @param a_char character to write
 */
static inline void
oswrch(uint8_t a_char)
{
	__asm__ __volatile__("jsr $ffee" : : "Aq" (a_char));
}

/**
 * OSRDCH Read character from currently selected input stream
 *
 * @returns character read
 */
static inline uint8_t
osrdch(void)
{
	uint8_t rega = 0;

	__asm__ __volatile__("jsr $ffe0" : "+Aq" (rega) );

	return rega;
}

/**
 * OSNEWL Write a newline to selected output stream
 */
static inline void
osnewl(void)
{
	__asm__ __volatile__("jsr $ffe7" : : : "a");
}

/**
 * OSASCI Write character routine where OSNEWL is called when A 0x0D
 *
 * This routine performs an OSWRCH call with the accumulator
 * contents unless called with 0x0D (13) when an OSNEWL call is
 * performed.
 *
 * @param a_char character to write
 */
static inline void
osasci(uint8_t a_char)
{
	__asm__ __volatile__("jsr $ffe3" : : "Aq" (a_char));
}

/**
 * GSINIT General string input initialise routine
 *
 * This call cannot be used from a co processor 
 */
static inline void
gsinit(uint8_t *a, uint8_t *y)
{
// TODO
}


/**
 * GSREAD Read character from string input
 *
 * This call cannot be used from a co processor 
 */
static inline void
gsread(uint8_t *a, uint8_t *x, uint8_t y)
{
// TODO
}

/**
 * OSRDRM Read byte in paged ROM
 *
 * This call cannot be used from a co processor
 *
 * @param y_romnum ROM bank number 0-15
 * @param address address in ROM, range 0x8000 - 0xbfff
 * @returns byte read
 */
extern uint8_t
osrdrm(uint8_t y_romnum, uintptr_t address);

/**
 * OSRDSC  Read a byte from current screen memory
 *
 * This call cannot be used from a co processor
 *
 * This call is synonumous with OSRDRM with the address in the range
 * of screen memory not paged ROM
 *
 * @param address address in screen memory
 * @returns byte read
 */
static inline uint8_t
osrdsc(uintptr_t address)
{
// UNTESTED
	return osrdrm(0, address);
}

/**
 * OSWRSC write byte to screen memory
 *
 * @param a_byte value to be written
 * @param address address to be written to
 */
extern void
oswrsc(uint8_t a_byte, uintptr_t address);

/**
 * OSEVEN Generate an event
 */
static inline void
oseven(uint8_t y_eventnum)
{
// TODO
}

/** 
 * OSCLI Passes line of text to the CLI
 *
 * @param xy_command 0 terminated command to pass to CLI
 */
static inline void
oscli(char *xy_command)
{
	uint8_t regx = ((uintptr_t) xy_command) & 0xff;
	uint8_t regy = (((uintptr_t) xy_command) & 0xff00) >> 8;

	/* convert from C zero terminator to BBC OS terminator */
	xy_command[strlen(xy_command)] = 0x0d;

	__asm__ __volatile__("jsr $fff7" : : "xq"(regx), "yq"(regy) );
}

/**
 * OSBYTE - OS call specified by the contents of A
 * taking parameters in X and Y
 *
 * @param a_func OS function to call
 * @param x parameter to OS function
 * @param y parameter to OS function
 */
static inline void
osbyte(uint8_t a_func, uint8_t *x, uint8_t *y)
{
	uint8_t regx = *x;
	uint8_t regy = *y;

	__asm__ __volatile__("jsr $fff4" : "+xq"(regx), "+yq"(regy) : "Aq" (a_func));
	*x = regx;
	*y = regy;
}

/**
 * OSWORD OS call specified by contents of A taking
 * parameters in a parameter block
 *
 * @param a_func OS function to call
 * @param xy_pointer pointer to parameter block
 */
extern void
osword(uint8_t a_func, const void *xy_pointer);

/**
 * Parameter block to be passed to osfile calls
 */
typedef struct osfile_parameterblock_s
{
	int32_t file_name:     16;  /**< Address of filename, terminated by 0x0D */
	int32_t load_address:  32;  /**< Load address of the file */
	int32_t exec_address:  32;  /**< Execution address of the file */
	int32_t start_address: 32;  /**< Start address of data for save, length of file otherwise */
	int32_t end_address:   32;  /**< End address of data for save, file attributes otherwise */
} osfile_parameterblock;

/**
 * Action to be performed by osfile call
 */ 
typedef enum osfile_action_e
{
	osfile_saveblock       = 0,
	osfile_writeinfo       = 1,
	osfile_writeload       = 2,
	osfile_writeexec       = 3,
	osfile_writeattr       = 4,
	osfile_readinfo        = 5,
	osfile_deletenamedfile = 6,
	osfile_loadnamedfile   = 0xff,
} osfile_action;

/**
 * OSFILE Read or write a whole file or its attributes
 * 
 *  @param a_action sub action osfile should perform
 * @param xy_pblock pointer to parameter block, also used for returning values
 */
extern void
osfile(uint8_t a_action, const osfile_parameterblock *xy_pblock);

/**
 * OSARGS Read or write an open file's arguments
 *
 * @param a_action number specifying the action required (also returns data)
 * @param x_zp_pointer points to a four byte zero page control block
 * @param y_filehandle file handle as provided by OSFIND, or zero 
 */
// TODO requires zero page access, how much does the C program mess with that?
// void osargs(uint8_t *a_action, uint8_t x_zp_poiner, uint8_t y_filehandle

/**
 * OSBGET Get one byte from an open file
 * 
 * @param y_handle the file handle, as provided by OSFIND
 */
static inline uint8_t
osbget(uint8_t y_handle)
{
	uint8_t rega = 0;

	__asm__ __volatile__("jsr $ffd7" : "+Aq" (rega) : "yq"(y_handle) );

	return rega;
}

/**
 * OSBPUT Write a single byte to an open file
 *
 * @param a_byte the byte to be written
 * @param y_handle the file handle, as provided by OSFIND
 */
static inline void
osbput(uint8_t a_byte, uint8_t y_handle)
{
	__asm__ __volatile__("jsr $ffd4" :  : "Aq" (a_byte), "yq"(y_handle) );
}

/**
 * Control block to be passed to osgbpb calls
 */ 
typedef struct osgbpb_controlblock_s
{
	int32_t file_handle: 8;   /**< file handle, as provided by OSFIND */
	int32_t data_pointer: 32; /**< TODO */
	int32_t num_bytes: 32;    /**< TODO */
	int32_t seq_pointer: 32;  /**< TODO */
} osgbpb_controlblock;

/**
 * Action to be performed by osgbpb call
 */
typedef enum osgbpb_action_e
{
	osgbpb_putbytes_seq       = 1,
	osgbpb_putbytes           = 2,
	osgbpb_getbytes_seq       = 3,
	osgbpb_getbytes           = 4,
	osgbpb_gettitleandbootopt = 5,
	osgbpb_readdirectory      = 6,
	osgbpb_readlibrary        = 7,
	osgbpb_readfilenames      = 8
} osgbpb_action;

/**
 * OSGBPB Read or write a group of bytes
 *
 * This routine transfers a number of bytes to or from an open file.
 * It can also be used to transfer filing system information.
 * 
 * @param a_action information to be transferred.
 * @param xy_pblock pointer to a control block in memory
 */
extern void
osgbpb(uint8_t a_action, const osgbpb_controlblock *xy_pblock);


/**
 * OSFIND Open or close a file for byte access
 * 
 * @param a_operation 0 = close, 0x40 = open for read, 0x80 = open for write, 0xc0 = open for read/write
 * @param x
 * @param y
 * @returns 
 */
extern uint8_t
osfind(uint8_t a_operation, uint8_t x, uint8_t y);

/**
 * Open a file for reading (OSFIND helper)
 *
 * @param filename 0 terminated filename (modifies the terminator)
 * @returns filehandle
 */
static inline uint8_t
osfind_open_r(char *filename)
{
	filename[strlen((filename))] = 0x0d;
	return osfind(0x40, ((uintptr_t) filename & 0xff), ((uintptr_t) filename & 0xff00) >> 8);
}

/**
 * Open a file for writing (OSFIND helper)
 *
 * @param filename 0 terminated filename (modifies the terminator)
 * @returns filehandle
 */
static inline uint8_t
osfind_open_w(char *filename)
{
	filename[strlen((filename))] = 0x0d;
	return osfind(0x80, ((uintptr_t) filename & 0xff), ((uintptr_t) filename & 0xff00) >> 8);
}

/**
 * Open a file for reading and writing (OSFIND helper)
 *
 * @param filename 0 terminated filename (modifies the terminator)
 * @returns filehandle
 */
static inline uint8_t
osfind_open_rw(char *filename)
{
	filename[strlen((filename))] = 0x0d;
	return osfind(0xc0, ((uintptr_t) filename & 0xff), ((uintptr_t) filename & 0xff00) >> 8);
}

/**
 * Close an open file (OSFIND helper)
 *
 * @param handle open file handle
 */
static inline void
osfind_close(uint8_t handle)
{
	osfind(0, 0, handle);
}


#endif // !__OS_H__


