/*
Copyright 2020 Peter Howkins

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the
"Software"), to deal in the Software without restriction, including
without limitation the rights to use, copy, modify, merge, publish,
distribute, sublicense, and/or sell copies of the Software, and to
permit persons to whom the Software is furnished to do so, subject to
the following conditions:

The above copyright notice and this permission notice shall be included
in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/
#include "os.h"
#include <stdio.h>
#include "support.h"

static void
oswrch_test(void)
{
	// simple character output tests
	oswrch('H');
	oswrch('e');
	oswrch('l');
	oswrch('l');
	oswrch('o');
	oswrch(',');
	oswrch(' ');
	oswrch('W');
	oswrch('o');
	oswrch('r');
	oswrch('l');
	oswrch('d');
	oswrch('!');
	osnewl();
}

static void
osasci_test(void)
{
	osasci('H');
	osasci('e');
	osasci('l');
	osasci('l');
	osasci('o');
	osasci(',');
	osasci(' ');
	osasci('A');
	osasci('g');
	osasci('a');
	osasci('i');
	osasci('n');
	osasci('!');
	osasci(0x0D);
}

static void
osrdch_test(void)
{
	// simple character input test
	puts("press a key\n");
	{
		uint8_t key = osrdch();
		puts("You pressed ");
		putchar(key);
		osnewl();
	}
}

static void
oscli_test(void)
{
	// Commandline interface
	oscli("INFO OSTEST");
}

static void
osbyte_test(void)
{
	// osbyte = *FX
	// Data is both passed in and out in the x and y registers, use pointers to values
	uint8_t x = 0;
	uint8_t y = 0;
	osbyte(0x75, &x, &y); // *FX 75 = Read VDU status

	putstring("VDU status = "); puth(x); osnewl();
}

static void
osword_test(void)
{
	// play a sound
	uint8_t block[8];

	block[0] = (uint8_t)    1;
	block[1] = (uint8_t)   (1 >> 8); // channel
	block[2] = (uint8_t)  -15;
	block[3] = (uint8_t) (-15 >> 8); // amplitude
	block[4] = (uint8_t)   53;
	block[5] = (uint8_t)  (53 >> 8); // pitch
	block[6] = (uint8_t)   20;
	block[7] = (uint8_t)  (20 >> 8); // duration

	osword(7, block);
}

static void
osfind_bput_bget_test(void)
{
	// File handling single bytes OSBPUT/OSBGET
	uint8_t file_handle;
	char filename[] = "testf";

	// Change the filename terminator from C /0 to BBC style
	filename[strlen(filename)] = 0x0d;

	// Open file for writing
	file_handle = osfind(0x80, (uintptr_t) filename & 0xff, ((uintptr_t) filename & 0xff00) >> 8);

	// write some sequential numbers
	osbput(0, file_handle);
	osbput(1, file_handle);
	osbput(2, file_handle);

	// close file
	osfind(0, 0, file_handle);

	// Open file for reading
	file_handle = osfind(0x40, (uintptr_t) filename & 0xff, ((uintptr_t) filename & 0xff00) >> 8);

	// read and check the numbers are sequential
	{
		uint8_t i, byte;

		for(i = 0; i <=2 ; i++) {
			byte = osbget(file_handle);

			if(byte != i) {
				puts("Error on read!");
			}
		}
	}

	// close file
	osfind(0, 0, file_handle);
}

static void
osfind_gbpb_test(void)
{
	// File handling multiple bytes OSGBPB
	uint8_t file_handle;
	char filename[] = "testf";
	uint8_t data[4] = { 0, 1, 2, 3 };
	osgbpb_controlblock block;

	// Change the filename terminator from C /0 to BBC style
	filename[strlen(filename)] = 0x0d;

	// Open file for writing
	file_handle = osfind(0x80, (uintptr_t) filename & 0xff, ((uintptr_t) filename & 0xff00) >> 8);

	// Setup the control block
	block.file_handle = file_handle;
	block.data_pointer = (intptr_t) &data;
	block.num_bytes = sizeof(data);
	block.seq_pointer = 0;

	osgbpb(osgbpb_putbytes, &block);

	// close file
	osfind(0, 0, file_handle);

	// Open file for reading
	file_handle = osfind(0x40, (uintptr_t) filename & 0xff, ((uintptr_t) filename & 0xff00) >> 8);

	// read and check the numbers are sequential
	{
		uint8_t i;
		uint8_t data2[4];

		// Setup the control block
		block.file_handle = file_handle;
		block.data_pointer = (intptr_t) &data2;
		block.num_bytes = 4;
		block.seq_pointer = 0;

		osgbpb(osgbpb_getbytes, &block);

		// Make sure bytes read = bytes written
		for(i = 0; i < 4 ; i++) {
			if(data[i] != data2[i]) {
				puts("Error on read!");
			}
		}
	}

	// close file
	osfind(0, 0, file_handle);
}

static void
osfind_helper_test(void)
{
	// File handling using helper functions for OSFIND
	uint8_t file_handle;
	uint8_t data[4] = { 0, 1, 2, 3 };
	osgbpb_controlblock block;

	// Open file for writing
	file_handle = osfind_open_w("testf");

	// Setup the control block
	block.file_handle = file_handle;
	block.data_pointer = (intptr_t) &data;
	block.num_bytes = sizeof(data);
	block.seq_pointer = 0;

	osgbpb(osgbpb_putbytes, &block);

	// close file
	osfind_close(file_handle);

	// Open file for reading
	file_handle = osfind_open_r("testf");

	// read and check the numbers are sequential
	{
		uint8_t i;
		uint8_t data2[4];

		// Setup the control block
		block.file_handle = file_handle;
		block.data_pointer = (intptr_t) &data2;
		block.num_bytes = 4;
		block.seq_pointer = 0;

		osgbpb(osgbpb_getbytes, &block);

		// Make sure bytes read = bytes written
		for(i = 0; i < 4 ; i++) {
			if(data[i] != data2[i]) {
				puts("Error on read!");
			}
		}
	}

	// close file
	osfind_close(file_handle);
}


static void
osfile_test(void)
{
	// OSFILE, whole file operations
	// Read info of file
	char filename[] = "testf";
	osfile_parameterblock pblock;

	// Change the filename terminator from C /0 to BBC style
	filename[strlen(filename)] = 0x0d;

	pblock.file_name = (intptr_t) filename;

	osfile(osfile_readinfo, &pblock);

	putstring("load = ");  putd(pblock.load_address);
	putstring(" exec = "); putd(pblock.exec_address);
	putstring(" len = ");  putd(pblock.start_address);
	putstring(" attr = "); putd(pblock.end_address); osnewl();

	pblock.file_name = (intptr_t) filename;

	// Delete file
	osfile(osfile_deletenamedfile, &pblock);
}

static void
osrdrm_test(void)
{
	char romname[17];
	uint8_t i = 0, cha, val;
	int8_t rom = 0;

	// Loop round each paged rom and output their title
	for(rom = 15; rom >= 0; rom--) {
		i = 0;
		for(cha = 9; cha <= 24; cha++) {
			val = osrdrm(rom, 0x8000 + cha);
			romname[i] = val;
			i++;
			if(val == '\0') {
				break;
			}
		}
		romname[16] = '\0';
		puth(rom);
		putstring(" ");
		putstring(romname);
		osnewl();
	}
}

static void
osrdsc_wrsc_test(void)
{
	uint8_t val = osrdsc(0x3000);

	// Should be 0 as screen has been cleared
	if(val != 0) {
		putstring("Error on screen read "); puth(val); osnewl();
	}

	// Set a byte to somethign obvious
	oswrsc(0xff, 0x3000);

	// Reread, should match what we set it to
	val = osrdsc(0x3000);

	// Should be 0 as screen has been cleared
	if(val != 0xff) {
		putstring("Error on screen read2 "); puth(val); osnewl();
	}

	// Make the first 4 character areas white
	uint8_t i;
	for(i = 0; i < 80; i++) {
		oswrsc(0xff, 0x3000 + i);
	}

	// move the cursor on so text doesn't splat this white block 
	osnewl();

}

int
main(void)
{
	// MODE 1
	oswrch(22);
	oswrch(1);

	osrdsc_wrsc_test();
	oswrch_test();
	osasci_test();
	osrdch_test();
	oscli_test();
	osbyte_test();
	osword_test();
	osfind_bput_bget_test();
	osfind_gbpb_test();
	osfind_helper_test();
	osfile_test();
	osrdrm_test();

	return 0;
}
